//##########################################################
//
// Project : Xchange version 1.24
//
// Author : Cyril MOTTIER
//
// Mail : orlin61@wanadoo.fr
//
// Site : www.cyrilmottier.com
//
// C Source File (main program for V200 and Ti92p)
//
// Created 08/05/2005 with TIGCC 0.96 beta 3 15:39:40
//
//###########################################################

#include <tigcclib.h>
#include "defines.h"
#include "gfx_data.h"
#include "prototypes.h"
#include "extgraph.h"

#define NO_EXIT_SUPPORT  //Smaller code
#define OPTIMIZE_ROM_CALLS  //Smaller code

#define PLANE1 GetPlane(LIGHT_PLANE)
#define PLANE2 GetPlane(DARK_PLANE)

//############################################################################
//#                                                                          #
//#                            VARIABLES DEFINITION                          #
//#                                                                          #
//############################################################################

// Initial highscores table
// These 5 variables will be saved in the external file
static HSC_ITEM bestscore[3]={{"Cyril",200},{"Cyril",200},{"Cyril",200}};// Mode 'Normal'
static HSC_ITEM bestmin[3]={{"Cyril",2376},{"Cyril",200},{"Cyril",200}};//Mode '5 Minutes'
static BT_ITEM bestfive[3]={{"Cyril",3,42,4},{"Cyril",4,0,0},{"Cyril",4,0,0}};// Mode 'Five Squares'
static int mode=0;
static int total_time[3]={0,0,0}; //total playing time... total_time[0] : hour, total_time[1] : minutes, total_time[2] : secondes

int (*map)[6]=NULL;
//int map[6][6];
int piece;
int add_to_score;
int joker;
unsigned short *fnt1 = NULL;
unsigned short *fnt2 = NULL;
//This buffer will be necessary to display the time and the menu 
unsigned char *my_f_buffer = NULL;
//unsigned char *sprt_save; //this pointer will be necessary for graphic fall
//unsigned char *big_mask;

INT_HANDLER ai_1 = NULL; // used to save the old interrupt 1 handler
INT_HANDLER ai_5 = NULL; //used to save the old interrupt 5 handler


//Clock Definition...
volatile int mseconds50 = 0; // = 50 milliseconds
volatile int seconds    = 0; // one second
volatile int dix    = 0;     // second/10=100 milliseconds
volatile int minutes    = 0;  // one minute
volatile int resettime  = 1;  // reset the clock
volatile int running  = 0;     // if 1,the clock is running...
volatile int mseconds50_2 = 0;

volatile int blinking = 0;
volatile int a_x_graph_pos = 0;
volatile int a_y_graph_pos = 0;
volatile int a_row = 0;
volatile int a_col = 0;
volatile int is_drawn = 1;
volatile int force_stop =0;


DEFINE_INT_HANDLER (myint5handler)
{
  if (blinking) //if a piece need to blink
  {
  	mseconds50_2++;
  	if (mseconds50_2==9)
  	{
  		if (is_drawn)
  		{
  			GraySprite8_AND(a_x_graph_pos, a_y_graph_pos, 8, mask, mask, PLANE1, PLANE2);
  			is_drawn=0;
  		}
  		else 
  		{
  			GraySprite8_OR(a_x_graph_pos, a_y_graph_pos, 8, game_sprt[map[a_row][a_col]][0], game_sprt[map[a_row][a_col]][1], PLANE1, PLANE2);
  			is_drawn=1;
  		}
  		mseconds50_2=0;
  	}
  }
  if (resettime)
  {
    mseconds50=seconds=minutes=resettime=0;
  }
  else if (running)
  {
    mseconds50++;
    dix=mseconds50>>1; // mseconds50 = 2 * dix
    if ((dix==2)||(dix==5)||(dix==8))
    {
    	sprintf(my_f_buffer, "%d'%d'%d", minutes, seconds, dix);
      GrayDrawRect2B(112, 87, 158, 95, COLOR_WHITE, RECT_FILLED, PLANE1, PLANE2);
      My_Gray_Font(134, 87, my_f_buffer, 6, PLANE1, PLANE2, fnt2, 1);
    }
    if (mseconds50 == 19) // 19 * mseconds50 = one second
    {   
      mseconds50 = 0;
      seconds++;
      if (seconds == 60)
      {
        seconds = 0;
        minutes++;
      }
      if (mode==1 && minutes==5) force_stop=1;
    }
  }
}


//############################################################################
//#                                                                          #
//#                               MISC FUNCTIONS                             #
//#                                                                          #
//############################################################################

//Add the current time to the total playing time
void Add_To_Total_Time (int m, int s)
{
	total_time[2]+=s;
	if (total_time[2]>59)
	{
		total_time[1]++;
		total_time[2]=total_time[2]%60;
	}
	total_time[1]+=m;
	if (total_time[1]>59)
  {
	  total_time[0]+=total_time[1]/60;
    total_time[1]=total_time[1]%60;
	}
}

//Normally we'll never have a m->col inferior to zero !!
int Is_A_Out_MAT_COORD (MAT_COORD *m)
{
	return (m->row>5);
}

int MAT_COORD_equals (MAT_COORD *a, MAT_COORD *b)
{
	return ((a->row==b->row)&&(a->col==b->col));
}

int Sum_From_One_To_Kp2 (int k)
{
	return (((k+(k==0 ? 0 : 2))*(k+3))>>1);
}

void StopBlinking (void)
{
	blinking=0;
  if (a_row<6) //If it's a valid MAT_COORD
    GraySprite8_OR(a_x_graph_pos, a_y_graph_pos, 8, game_sprt[map[a_row][a_col]][0], game_sprt[map[a_row][a_col]][1], PLANE1, PLANE2);	
}

//Check if the current time is before than the time passed in argument
short Is_Before (short m, short s, short d)
{
	return ((m*600+s*10+d)>(minutes*600+seconds*10+dix));
}

/*===========================================================================*/
/* "waits" a given number of milliseconds (dummy looping)                    */
/* NOTE: this wait loop is parametrized to fit on HW1 calculators            */
/*===========================================================================*/
//Thanks to Thomas Nussbaumer for this routine extracted from Ti-Chess source
asm("xdef WaitForMillis\n"
"WaitForMillis:  movem.l %d0-%d3,-(%sp)\n"
"           move.w (20,%sp),%d2\n"
"           move.l #31,%d1\n"
"           move.l #31,%d3\n"
"_wl2_:     move.w #120,%d0    /* modify this value for exact timing !!! */\n"
"_wl1_:     rol.l  %d3,%d1\n"
"           dbra   %d0,_wl1_\n"
"           dbra   %d2,_wl2_\n"
"           movem.l (%sp)+,%d0-%d3\n"
"           rts");

void Wait_User_Action(void)
{
	WaitForMillis(500);
	for (;;)
  {
    BEGIN_KEYTEST
      if (_keytest_optimized (RR_ENTER1)
       || _keytest_optimized (RR_ENTER2)
       || _keytest_optimized (RR_ENTER)
       || _keytest_optimized (RR_ESC))
        break;
    END_KEYTEST
  }
}

//Key is a simple function which return a code depending on 
//pressed key. It's a sort of ngetchx() but with _rowread
short key(short time)
{
short value=0, pseudo_s=0, wait_value=time*HW;
while (value==0)
 {
	 BEGIN_KEYTEST
	 // [ESC] is pressed
   if 	(_keytest_optimized(RR_ESC) && pseudo_s>wait_value) value=ESC;
   // [ENTER] is pressed
   if 	((_keytest_optimized(RR_ENTER1) || _keytest_optimized(RR_ENTER2)) && pseudo_s>wait_value) value=ENTER;
   // [CLEAR] is pressed
   if (_keytest_optimized(RR_CLEAR) && pseudo_s>wait_value) off();
   // [+] is pressed
   if (_keytest_optimized(RR_PLUS) && pseudo_s>wait_value) {OSContrastUp(); value=CNTR_UP;}
   // [-] is pressed
   if (_keytest_optimized(RR_MINUS) && pseudo_s>wait_value) {OSContrastDn(); value=CNTR_DOWN;}
   // [LEFT] is pressed
   if (_keytest_optimized(RR_LEFT) && pseudo_s>wait_value) value=LEFT;
   // [DOWN] is pressed
   if (_keytest_optimized(RR_DOWN) && pseudo_s>wait_value) value=DOWN;
   // [RIGHT] is pressed
   if (_keytest_optimized(RR_RIGHT) && pseudo_s>wait_value) value=RIGHT;
   // [UP] is pressed
   if (_keytest_optimized(RR_UP) && pseudo_s>wait_value) value=UP;
   if (pseudo_s<(wait_value+2)) pseudo_s++;
   if (force_stop) value=MAGIC_KEY;
   END_KEYTEST
 }
return value;
}

//############################################################################
//#                                                                          #
//#                            HIGHSCORE FUNCTIONS                           #
//#                                                                          #
//############################################################################

// Here is the structure of the HSC file :
// 'Normal' highscores
// '5 Minutes' highscores
// '5 Squares' highscores
// The mode
// The total playing time

void Open_HSC(void)
{
	FILE *hsc_file;
	EM_moveSymFromExtMem (SYMSTR ("xchsave") ,HS_NULL);	// unarchive the file if it necessary
	hsc_file = fopen("xchsave", "rb");
	// Check if the file had been correctly opened and if
	// the 'magic number' is right (equals to HSC_FILE_CODE)
	if (hsc_file != NULL && (fgetc(hsc_file)==HSC_FILE_CODE))
	{
		//Then it reads all datas in the file
		fread(bestscore, sizeof(HSC_ITEM), 3, hsc_file);
		fread(bestmin, sizeof(HSC_ITEM), 3, hsc_file);
		fread(bestfive, sizeof(BT_ITEM), 3, hsc_file);
		fread(&mode, sizeof(int), 1, hsc_file);
		fread(total_time, 3*sizeof(int), 1, hsc_file);
	}
	//if the file doesn't exist or is a fake (create/overwrite it)
	else  hsc_file = fopen("xchsave", "wb");
	fclose(hsc_file);	// close the file
}


void Close_HSC(void)
{
  FILE *hsc_file;
  hsc_file = fopen("xchsave", "wb");	// open the file
  fputc(HSC_FILE_CODE, hsc_file);	// first of all we write the magic number at the beginnig of our file
  fwrite(bestscore, sizeof(HSC_ITEM), 3, hsc_file);	// copy option
  fwrite(bestmin, sizeof(HSC_ITEM), 3, hsc_file); // copy highscores (score)
  fwrite(bestfive, sizeof(BT_ITEM), 3, hsc_file); // copy highscores (time)
  fwrite(&mode, sizeof(int), 1, hsc_file);
  fwrite(total_time, sizeof(int), 3, hsc_file);
	fputc(0,hsc_file);
  fputs("HSC",hsc_file);
  fputc(0,hsc_file);
  fputc(OTH_TAG,hsc_file);
	fclose(hsc_file);	// close the file
	EM_moveSymToExtMem (SYMSTR ("xchsave") ,HS_NULL);	// archive the file
}

void Input_HSC_Str(unsigned char *buffer)
{
  GKeyFlush();
  GraySetInt1Handler(ai_1);
	void *kbq = kbd_queue();
	unsigned short int key, i=0;
	
	memset(PLANE1, 0, LCD_SIZE);
	memset(PLANE2, 0, LCD_SIZE);
	
	DrawGrayStrExt(0, 23, "CONGRATULATIONS !!!", A_NORMAL | A_CENTERED, F_8x10);
	DrawGrayStrExt(0, 43, "New highscore!", A_NORMAL | A_CENTERED, F_6x8);
	DrawGrayStrExt(0, 53, "Enter your name:", A_NORMAL | A_CENTERED, F_6x8);
	DrawGrayStrExt(115, 73, "_", A_NORMAL, F_6x8);
	DrawGrayStrExt(0, 95, "You can send your highscore to", A_NORMAL | A_CENTERED, F_4x6);
	DrawGrayStrExt(0, 103, "orlin61@wanadoo.fr", A_NORMAL | A_CENTERED, F_4x6);

	while (TRUE)
	{
		if (!OSdequeue(&key,kbq))
		{
			if (((key==KEY_ENTER) || (key==KEY_ESC)) && (i))
			{
				GraySetInt1Handler(DUMMY_HANDLER);
				buffer[i]=0;
				return;
			}
		  //check if the char can be printed...
			if (key >= ' ' && key <= '~' && i!=8) buffer[i++]=key;
			//check if you want to erase your last letter
			else if (key==KEY_BACKSPACE && i) i--;
			
			if (i==8) buffer[i]=0;
			else 
			{
			  buffer[i]='_';
			  buffer[i+1]=0;
			}
			//draw your name in the center of the screen
			DrawGrayRect(69 , 73 , 169 , 81 , COLOR_WHITE , RECT_FILLED);
			DrawGrayStrExt(0, 73, buffer, A_NORMAL | A_CENTERED, F_6x8);
		}
	}
}

void Change_HSC_Table (unsigned long score)
{
  Add_To_Total_Time(minutes,seconds);
  //If "Normal"
  if (mode==0)
  {
    //Rank records
  	if (bestscore[2].score<score)
  	{
	  	int a;
	  	for(a=2;(bestscore[a-1].score<score)&&a;a--) memcpy(&bestscore[a],&bestscore[a-1],sizeof(HSC_ITEM));
      bestscore[a].score=score;
      Input_HSC_Str(bestscore[a].name);
	  }
  }
  
  //If "5 Minutes"
  if (mode==1)
  {
    //Rank records
  	if (bestmin[2].score<score)
  	{
	  	int a;
	  	for(a=2;(bestmin[a-1].score<score)&&a;a--) memcpy(&bestmin[a],&bestmin[a-1],sizeof(HSC_ITEM));
      bestmin[a].score=score;
      Input_HSC_Str(bestmin[a].name);
	  }
  }
  
  //If "Five Squares"
  if (mode==2)
  {
    //Rank records
  	if (Is_Before (bestfive[2].min, bestfive[2].sec, bestfive[2].dix))
  	{
	  	int a;
	  	for(a=2;Is_Before (bestfive[a-1].min, bestfive[a-1].sec, bestfive[a-1].dix)&&a;a--)
	  	  memcpy(&bestfive [a],&bestfive[a-1],sizeof(BT_ITEM));
      bestfive[a].min=minutes;
      bestfive[a].sec=seconds;
      bestfive[a].dix=dix;
      Input_HSC_Str(bestfive[a].name);
	  }
  }
}


//############################################################################
//#                                                                          #
//#                             GRAPHIC FUNCTIONS                            #
//#                                                                          #
//############################################################################

short My_Title (void)
{
	const char *about[5]={"Xchange v1.24", "Coded by Cyril MOTTIER", "on November 2005", "orlin61@wanadoo.fr", "www.cyrilmottier.com"};
	const char *titles[4]={"Game", "Highscores", "About", "Quit"};
	const char *game[3]={"Normal", "5 Minutes", "5 Squares"};
	short n_title=0, n_game, pressed_key, x;
	unsigned char *pic1 = RLE_Decompress_To_ExtGraph (title_l, 855);
	unsigned char *pic2 = RLE_Decompress_To_ExtGraph (title_d, 855);
	
	if (!(pic1&&pic2)) return 0;
	
	memset(PLANE1, 0, LCD_SIZE);
	memset(PLANE2, 0, LCD_SIZE);
	GraySpriteX8_OR(43, 17, 45, pic1, pic2, 19, PLANE1, PLANE2); 
  free(pic2);
	free(pic1);
	
	start:
  n_title--;pressed_key=RIGHT;
	
	DrawGrayRect(39, 62, 198, 114, COLOR_WHITE, RECT_FILLED);
	GraySprite16_OR(58, 83, 16, sprt16[0][0], sprt16[0][1], PLANE1, PLANE2);
  GraySprite16_OR(164, 83, 16, sprt16[1][0], sprt16[1][1], PLANE1, PLANE2);
  DrawGrayStrExt(0, 68, "::Menu::", A_NORMAL | A_CENTERED, F_8x10);
  sprintf(my_f_buffer, "Total playing time : %d'%d'%d", total_time[0], total_time[1], total_time[2]);
  DrawGrayStrExt(0, 107, my_f_buffer, A_REPLACE | A_CENTERED, F_4x6);
	
	while (pressed_key!=ESC)
	{
    if (pressed_key==RIGHT && n_title<3)
    {
    	DrawGrayRect(79, 86, 159, 95, COLOR_WHITE, RECT_FILLED);
    	My_Gray_Font(119, 86 , titles[++n_title], 8, PLANE1, PLANE2, fnt1, 1);
    }
	  if (pressed_key==LEFT && n_title>0)
	  {
	  	DrawGrayRect(79, 86, 159, 95, COLOR_WHITE, RECT_FILLED);
	  	My_Gray_Font(119, 86 , titles[--n_title], 8, PLANE1, PLANE2, fnt1, 1);
	  } 
	  if (pressed_key==ENTER)
	  {
	 	  pressed_key=MAGIC_KEY;
	 	  if (n_title==3) return 0;
	    DrawGrayRect(39, 62, 198, 112, COLOR_WHITE, RECT_FILLED);
  	  sprintf (my_f_buffer, "::%s::", titles[n_title]);
      DrawGrayStrExt(0, (n_title ? 63 : 68), my_f_buffer, A_REPLACE | A_CENTERED, F_8x10);
	 	  if (n_title==0)
	 	  {
	 	    GraySprite16_OR(58, 83, 16, sprt16[0][0], sprt16[0][1], PLANE1, PLANE2);
	 	    GraySprite16_OR(164, 83, 16, sprt16[1][0], sprt16[1][1], PLANE1, PLANE2);
	 	    My_Gray_Font(119, 86 , game[mode], 8, PLANE1, PLANE2, fnt1, 1);
	 	    sprintf(my_f_buffer, "Total playing time : %d'%d'%d", total_time[0], total_time[1], total_time[2]);
        DrawGrayStrExt(0, 107, my_f_buffer, A_REPLACE | A_CENTERED, F_4x6);
	      while (pressed_key!=ESC)
        {
		      if (pressed_key==RIGHT && mode<2)
		      {
		      	DrawGrayRect(79, 86, 159, 95, COLOR_WHITE, RECT_FILLED);
		      	My_Gray_Font(119, 86 , game[++mode], 8, PLANE1, PLANE2, fnt1, 1);
		      }
	        if (pressed_key==LEFT && mode>0)
	        {
	        	DrawGrayRect(79, 86, 159, 95, COLOR_WHITE, RECT_FILLED);
	        	My_Gray_Font(119, 86 , game[--mode], 8, PLANE1, PLANE2, fnt1, 1);
	        }
	        if (pressed_key==ENTER) return 1;
	        pressed_key=key(200);
      	}
      	goto start;
	  	}
	  	if (n_title==1)
	 	  {
	 		  n_game=-1;
	 		  pressed_key=RIGHT;
	 		  GraySprite16_OR(63, 89, 16, sprt16[0][0], sprt16[0][1], PLANE1, PLANE2);
        GraySprite16_OR(159, 89, 16, sprt16[1][0], sprt16[1][1], PLANE1, PLANE2);
         while (pressed_key!=ESC)
        {
		      if (pressed_key==RIGHT && n_game<2)
		      {
		      	DrawGrayRect(81, 76, 158, 112, COLOR_WHITE, RECT_FILLED);
		      	DrawGrayStrExt(0, 76, game[++n_game], A_REPLACE | A_CENTERED, F_6x8);
		      	for (x=2;x>=0;x--)
		      	{
		      		if (n_game==2) sprintf(my_f_buffer, "%d. %s :: %hd'%hd'%hd", x+1, bestfive[x].name, bestfive[x].min, bestfive[x].sec, bestfive[x].dix);
		      		if (n_game==0) sprintf(my_f_buffer, "%d. %s :: %ld", x+1, bestscore[x].name, bestscore[x].score);
		      		if (n_game==1) sprintf(my_f_buffer, "%d. %s :: %ld", x+1, bestmin[x].name, bestmin[x].score);
		      		GrayDrawStrExt2B(0, 87+8*x, my_f_buffer, A_REPLACE | A_CENTERED, F_4x6, PLANE2, PLANE2);
		      	}
		      }
	        if (pressed_key==LEFT && n_game>0)
	        {
	        	DrawGrayRect(81, 76, 158, 112, COLOR_WHITE, RECT_FILLED);
		      	DrawGrayStrExt(0, 76, game[--n_game], A_REPLACE | A_CENTERED, F_6x8);
		      	for (x=2;x>=0;x--)
		      	{
		      		if (n_game==2) sprintf(my_f_buffer, "%d. %s :: %hd'%hd'%hd", x+1, bestfive[x].name, bestfive[x].min, bestfive[x].sec, bestfive[x].dix);
		      		if (n_game==0) sprintf(my_f_buffer, "%d. %s :: %ld", x+1, bestscore[x].name, bestscore[x].score);
		      		if (n_game==1) sprintf(my_f_buffer, "%d. %s :: %ld", x+1, bestmin[x].name, bestmin[x].score);
		      		GrayDrawStrExt2B(0, 87+8*x, my_f_buffer, A_REPLACE | A_CENTERED, F_4x6, PLANE2, PLANE2);
		      	}
	        }
	        if (pressed_key==ENTER) break;
	        pressed_key=key(200);
      	}
      	goto start;
	 	  }
	   	if (n_title==2)
	   	{
	 	  	for (x=0;x<5;x++) DrawGrayStrExt2B(0, 76+7*x, about[x], A_NORMAL | A_CENTERED, F_4x6, PLANE2, PLANE2);
	 	  	Wait_User_Action();
	 	  	goto start;
	  	}
	  }
	  pressed_key=key(200);
	}
return 0;
}


//This function decompressed a RLE_compressed picture from src to 
//a new allocated block : This function return a pointer to the decompressed
//picture. I have written a tutorial (in english) dealing with this problem so take a 
//tour to my web site : www.cyrilmottier.com
unsigned char *RLE_Decompress_To_ExtGraph (unsigned char *src, unsigned short size)
{
 unsigned char len, val, *ptr, *dest;
 ptr = malloc (size);
 if (ptr)
 {
  dest=ptr;
  while (size)
  {
   if (*src==0x91)
   {
    src++;
    val = *src++;
    len = *src++;
    while (len) 
    {
     *dest++ = val;
     len--;
     size--;
    }
   }
   else
   {
    *dest++ = *src++;
    size--;
   }
  }
 }
 return ptr;
}

void *IniFont (short taille)
{ 
  short i; 
  unsigned short *charset=NULL; 
  unsigned short *charset1=NULL; 
  unsigned short *charset3=NULL; 
  unsigned short *charset2=NULL;
   
  if (!(charset = malloc(6016))) return NULL; 
  SetCurAttr(A_NORMAL); 
	FontSetSys(taille); 
  memset(charset,0,6016); 
  PortSet(charset,15,15000); 
  for (i=32;i<126;i++) DrawChar(2,(i-32)<<4,(char)i,A_REPLACE); 
  PortRestore(); 
  if (!(charset2 = malloc(6016)))
  {
    free(charset);
    return NULL;
  }     
  charset1=charset2; 
  charset3=charset; 
  for (i=0;i<1504;i++) 
  {  
	  (*charset1)=((( ((*charset)<<1) | ((*charset)>>1)) ) |(*(charset-1))|(*(charset+1)))|(*charset);	 
	  (*(charset1+1504))= (((( ((*charset)<<1) | ((*charset)>>1)) ) |(*(charset-1))|(*(charset+1)))^(*charset))&(~(*charset)); 
	  charset1++; 
		charset++;			 
	}     
  free(charset3);  
  return charset2; 
} 
 
//If centered=1 the word is center on x
void My_Gray_Font (short x, short y,const unsigned char *text,short space,void *plane1, void *plane2, void *charset2, short centered) 
{ 
  short x2 = x;
  if (centered) x2=x-((strlen(text)*space)>>1);
  
  while (*text) 
	{ 
		if ((*text)!=32)   
		{ 
		Sprite16_OR(x2,y,10,charset2+32*((*text)-32)-2,plane1);	 
		Sprite16_OR(x2,y,10,charset2+32*((*text++)-32)-2+3008,plane2);	 
		} 
		else text++; 
		x2+=space; 
	} 
} 

void New_Square()
{
	int a,b;
	for (a=0;a<6;a++)
	{
	  for (b=0;b<6;b++)
	  {
	  	map[a][b]=random (3)+2;
	  	GraySprite8_MASK((b<<3)+2, (a<<3)+50, 8, game_sprt[map[a][b]][0], game_sprt[map[a][b]][1], mask, mask, PLANE1, PLANE2);
	  }
	}
}

void Draw_Square()
{
	int a,b;
	for (a=0;a<6;a++)
	  for (b=0;b<6;b++)
	  	GraySprite8_MASK((b<<3)+2, (a<<3)+50, 8, game_sprt[map[a][b]][0], game_sprt[map[a][b]][1], mask, mask, PLANE1, PLANE2);
}
	
//Function to draws the game set randomly and prepare the 2-dimensionnal array
//with right numbers in it
void Init_Gameset (void)
{
  memset(PLANE1, 0, LCD_SIZE);
	memset(PLANE2, 0, LCD_SIZE);
  GrayDrawRect(0, 0, 109, 99, COLOR_BLACK, RECT_EMPTY);
  GrayDrawRect(111, 0, 159, 99, COLOR_BLACK, RECT_EMPTY);
  
  GrayFastDrawHLine2B(111, 159, 33, COLOR_BLACK, PLANE1, PLANE2);
  GrayFastDrawHLine2B(111, 159, 67, COLOR_BLACK, PLANE1, PLANE2);
  
  My_Gray_Font(111, 6 , "Score:", 8, PLANE1, PLANE2, fnt1, 0);
  My_Gray_Font(111, 40 , "Lives:", 8, PLANE1, PLANE2, fnt1, 0);
  My_Gray_Font(115, 74 , "Time:", 8, PLANE1, PLANE2, fnt1, 0);
  int i,j;
  //The 'reflection' area is displayed
  for (i=0;i<5;i++)
    for (j=0;j<5-i;j++) GraySprite8_OR((i<<3)+2, (j<<3)+2, 8, game_sprt[1][0], game_sprt[1][1], PLANE1, PLANE2);
	New_Square();
}

//This is the principal function of this game because it's move graphicly and 
//matrix-caly the piece

void Move_Piece(int h)
{
  //if side=1 => the piece will arrive on the top of the square
	int side=0, num, x_pos, y_pos, f_x_graph_pos, f_y_graph_pos, piece_eaten=0, nothing_to_change=0, is_done_y=0, is_done_x=0, temp;
	MAT_COORD f_pos, a_pos;
	a_pos = Attack_Position(h);
	f_pos = Final_Position(h);
	
	if (h>=2 && h<=42)
	{
		num=5-((h-2)>>3);
		side=1;
	}
	else num=(h-50)>>3;
	
	if (!MAT_COORD_equals(&f_pos, &a_pos)) piece_eaten = 1;
	
	if (Is_A_Out_MAT_COORD(&f_pos)) //If the piece goes out of the square
	{
		f_x_graph_pos = (((f_pos.col)<<3)+2);
		f_y_graph_pos = (((f_pos.row)<<3)+42);
		nothing_to_change = 1;
	}
	else
	{
		f_x_graph_pos = (((f_pos.col+((!side && f_pos.row==num && !piece_eaten) ? 1 : 0))<<3)+2);
	 	f_y_graph_pos = (((f_pos.row+((!piece_eaten) ? -1 : 0))<<3)+50);
	}
		
		for (x_pos=99;x_pos>=f_x_graph_pos;x_pos--)
		{
			GraySprite8_MASK(x_pos, h, 8, game_sprt[piece][0], game_sprt[piece][1], mask, mask, PLANE1, PLANE2);
			FastDrawGrayLine(x_pos+8, h, x_pos+8, h+7, COLOR_WHITE);
			//Waiting loop
			WaitForMillis(HW);
		}
		for (y_pos=h+1;y_pos<=f_y_graph_pos;y_pos++)
		{
			GraySprite8_MASK(f_x_graph_pos, y_pos, 8, game_sprt[piece][0], game_sprt[piece][1], mask, mask, PLANE1, PLANE2);
			FastDrawGrayLine(f_x_graph_pos, y_pos-1, f_x_graph_pos+7, y_pos-1, COLOR_WHITE);
			//Waiting loop
			WaitForMillis(HW);
		}
		//Ici on affiche pendant 8 loops le sprite de changement
		for (y_pos=f_y_graph_pos-1;y_pos>=h;y_pos--)
		{
			if (f_y_graph_pos-y_pos<9 && !nothing_to_change && piece_eaten)
			{
				if (piece_eaten && !is_done_y)
				{
					temp = map[f_pos.row][f_pos.col];
  	      map[f_pos.row][f_pos.col]=piece;
  	      piece=temp;
  	      is_done_y=1;
				}
				GraySprite8_MASK(f_x_graph_pos, f_y_graph_pos, 8, game_sprt[map[f_pos.row][f_pos.col]][0], game_sprt[map[f_pos.row][f_pos.col]][1], mask, mask, PLANE1, PLANE2);
			}
			else
			{
				FastDrawGrayLine(f_x_graph_pos, y_pos+8, f_x_graph_pos+7, y_pos+8, COLOR_WHITE);
			}
			GraySprite8_MASK(f_x_graph_pos, y_pos, 8, game_sprt[piece][0], game_sprt[piece][1], mask, mask, PLANE1, PLANE2);
			//Waiting loop
			WaitForMillis(HW);
		}
		for (x_pos=f_x_graph_pos+1;x_pos<=100;x_pos++)
		{
			if (x_pos-f_x_graph_pos<9 && !nothing_to_change && !is_done_y && piece_eaten)
			{
				if (piece_eaten && !is_done_x)
				{
					temp = map[f_pos.row][f_pos.col];
  	      map[f_pos.row][f_pos.col]=piece;
  	      piece=temp;
  	      is_done_x=1;
				}
				GraySprite8_MASK(f_x_graph_pos, f_y_graph_pos, 8, game_sprt[map[f_pos.row][f_pos.col]][0], game_sprt[map[f_pos.row][f_pos.col]][1], mask, mask, PLANE1, PLANE2);
			}
			else
			{
				FastDrawGrayLine(x_pos-1, h, x_pos-1, h+7, COLOR_WHITE);
			}
			
			GraySprite8_MASK(x_pos, h, 8, game_sprt[piece][0], game_sprt[piece][1], mask, mask, PLANE1, PLANE2);
			//Waiting loop
			WaitForMillis(HW);
		}
	Fall();
	Draw_Square();
	/*Sprite8Get(2, 50, 32, PLANE1, sprt_save); //We are saving the lightplane
	Sprite8Get(2, 50, 32, PLANE2, sprt_save+48); //and the darkplane
	GraySprite8_OR(100, 10, 32, sprt_save, sprt_save+48, PLANE1, PLANE2);*/
	//Gfx_Fall();
}

void My_Window(const char *str)
{
	short str_width = strlen(str)<<3;
	short x0 = 70-(str_width>>1);//(160-str_width)/2-10
	short x1 = 90+(str_width>>1);//(160-str_width)/2+str_width+10
	GrayDrawRect2B(x0, 36, x1, 64, COLOR_BLACK, RECT_EMPTY, PLANE1, PLANE2);
	GrayDrawRect2B(x0+1, 37, x1-1, 63, COLOR_DARKGRAY, RECT_EMPTY, PLANE1, PLANE2);
	GrayDrawRect2B(x0+2, 38, x1-2, 62, COLOR_LIGHTGRAY, RECT_EMPTY, PLANE1, PLANE2);
	GrayDrawRect2B(x0+3, 39, x1-3, 61, COLOR_WHITE, RECT_FILLED, PLANE1, PLANE2);
	GrayDrawStrExt2B(x0+10, 46, str, A_NORMAL, F_8x10, PLANE1, PLANE2);
	Wait_User_Action();
}

//In this function arg[6][2] is an array which contains for each
//columns first the step (arg[col][0]) and then the cell where the fall starts
//(arg[col][1])
/*void Gfx_Fall (void)
{
	int col, row, step, i, stop, max=0;
	
	//First of all this function analyse the square to 'know' which columns
	//need to be lowered : it stores the values in an array of FOR_GFX_FALL
	FOR_GFX_FALL tab[6];
	for (col=0;col<6;col++)
	{
		stop=0;
		for (row=5;map[row][col]!=0&&!stop;row--);
    {
      if (Column_Sum(row,col)!=0)
      {
        step=1;
        i=row-1;
        while(map[i][col]==0)
        {
          step++;
          i--;
        }
        tab[col]=(FOR_GFX_FALL){i, step};//On enregistre dans i la ligne infrieure a ce qu'il faut baisser
        stop=2;
        if (step>max) max=step;
      }
      else stop=1;
    }
    if (stop!=2) tab[col]=(FOR_GFX_FALL){0, 0};
	}
	//The array pointed bu 'sprt_save' is filled with sprites
	for (col=0;col<6;col++)
	{
		Sprite8Get((col<<3)+2, 50, (tab[col].num_row+1)<<3, PLANE1, sprt_save+(48*col)); //We are saving the lightplane
		Sprite8Get((col<<3)+2, 50, (tab[col].num_row+1)<<3, PLANE2, sprt_save+(48*(col+1))); //and the darkplane
	}
	//GraySpriteX8_OR(100, 10, tab[col].num_row<<3, sprt_save+(48*col), sprt_save+(48*col+48), 1, PLANE1, PLANE2);
	//And then rows are lowered
	for (i=0;i<=max<<3;i++)
	{
		for (col=0;col<6;col++)
		{
			if (tab[col].cell_to_lower!=0)
			{
				GraySprite8_MASK((col<<3)+2, 50+i, (tab[col].num_row+1)<<3, sprt_save+(48*col), sprt_save+(48*(col+1)), big_mask, big_mask, PLANE1, PLANE2);
				GrayFastDrawLine((col<<3)+2, i-1, (col<<3)+10, i-1, COLOR_WHITE);
			}
		}
		WaitForMillis(HW*2);
	}
}*/

//############################################################################
//#                                                                          #
//#                            MATRIX FUNCTIONS                              #
//#                                                                          #
//############################################################################

//This function calculates the sum of all number you can find
//on the column 'column' and on top of the row 'row'
//This routine seems to be okay
short Column_Sum (short row, short column)
{
	short s=0,i;
	for (i=row-1;i>=0;i--) s+=map[i][column];
	return s;
}


//This function modifies the matrix by falling all colums which contains a zero (or more)
//for exemple the array :
//{4,4,0,0,0,0}
//{2,1,0,2,0,3}
//{1,2,4,2,1,4}
//{0,0,0,0,0,0}
//{0,3,3,2,1,2}
//{3,2,4,1,2,3}
//becomes :
//{0,0,0,0,0,0}
//{0,4,0,0,0,0}
//{4,1,0,2,0,3}
//{2,2,4,2,1,4}
//{1,3,3,2,1,2}
//{3,2,4,1,2,3}
//This routine seems to be okay 
void Fall (void)
{
  short row,col,i,step;
  for (col=0;col<6;col++)
  {
    for (row=5;row>0;row--)
    {
      //if the cell is filled this a zero, the column is lowered
      if (map[row][col]==0 && Column_Sum(row,col)!=0)
      {
        step=1;
        i=row-1;
        while(map[i][col]==0)
        {
          step++;
          i--;
        }
        for (i=row-step;i>=0;i--)
        {
          map[i+step][col]=map[i][col];
          map[i][col]=0;
        }
      }
    }
  }
}

//This function tries to find the number 'piece' and returns 0 (you are not a winner) if
//this number had been found else it returns 1 (you win!)
//This routine seems to be okay
int Is_A_Winner(void)
{
 int i,j;
	for (i=5;i>=0;i--)
	{
		for (j=0;j<6;j++)
		{
			if (map[i][j]==piece) return 0;
		}
	}
	return 1;
}


//This function returns 1 if you lose and 0 else.
//This function seems to be okay
int Is_A_Loser (void)
{
	short i,j;
	for (i=0;i<6;i++)
	{
		j=-1;
		do
		{
			j++;
			if (map[j][i]==piece) return 0;
		} while (map[j][i]==0 && j<5);
		
		j=-1;
		do
		{
			j++;
			if (map[i][5-j]==piece) return 0;
		} while (map[i][5-j]==0 && j<5);
		
	}
	return 1;
}

//It finds the attack position in the square
//This function seems to be okay
MAT_COORD Attack_Position(int h)
{
	//if side=1 => the piece will arrive on the top of the square
	int side=0, num, row=0, col=5;
	if (h>=2 && h<=42)
	{
		num=5-((h-2)>>3);
		side=1;
	}
	else num=(h-50)>>3;
	
	if(!side)
	{
		while (map[num][col]==0 && col>-1)
		{
		  col--;
			if (col==-1) {side=1; num=0;}
		}
		if (!side) return (MAT_COORD){num, col};
	}
	
	if (side)
	{
		while (map[row][num]==0 && row<6) row++;
		return (MAT_COORD){row, num};
	}
}

//This function is an essential and important routine because it returns a MAT_COORD structure
//which represents the finish position of the piece in the square (map). First of all it calculates 
//the position of the piece depending on 'h' and then it search the finish position
//Note that if the it's fill the cell with zero if some pieces in the square are 'eaten'
//This function seems to be okay
MAT_COORD Final_Position(int h)
{
	//if side=1 => the piece will arrive on the top of the square
	int side=0, num, row=0, col=5;
	add_to_score=0;
	if (h>=2 && h<=42)
	{
		num=5-((h-2)>>3);
		side=1;
	}
	else num=(h-50)>>3;
	
	if(!side)
	{
		while ((map[num][col]==0 || map[num][col]==piece) && col>-1)
		{
		  if (map[num][col]!=0) add_to_score++;
		  map[num][col]=0;
		  col--;
			if (col==-1) {side=1; row=num+1; num=0;}
		}
		if (!side) return (MAT_COORD){num, col};
	}
	
	if (side)
	{
		while ((map[row][num]==0 || map[row][num]==piece) && row<6)
		{
		  if (map[row][num]!=0) add_to_score++;
		  map[row][num]=0;
		  row++;
	}
		return (MAT_COORD){row, num};
	}
}

//############################################################################
//#                                                                          #
//#                               MAIN FUNCTION                              #
//#                                                                          #
//############################################################################

void _main(void)
{	
  //First we text if there is enough free memory to launch the game
  unsigned char *test;
  if (!(test = malloc(50000))) goto end;
  else free(test);
  
  //Initialize fonts, buffer and variables
  my_f_buffer=malloc(30);
  map=calloc(6,sizeof(*map));
  fnt1 = IniFont(F_6x8);
  fnt2 = IniFont(F_4x6);
  unsigned long score;
  int lives, y_piece, pressed_key, nb_f_squares;
  char my_buffer[10];
  
  //Save screen
  LCD_BUFFER screen;
  LCD_save(screen);
  
  //save interrupt handler
  ai_1 = GetIntVec (AUTO_INT_1);
  ai_5 = GetIntVec (AUTO_INT_5);
  SetIntVec (AUTO_INT_1, DUMMY_HANDLER);
  SetIntVec(AUTO_INT_5, myint5handler);
  
  //Open highscores
  Open_HSC();
  //if (!(sprt_save = malloc(624))) goto end; //624 = 6*6*8*2 (36 sprites of 8 bytes for 2 planes) + 48 (blank sprite)
  //big_mask = sprt_save+576;
  //memset(sprt_save, 0, 624);
  
  if (!GrayOn()) goto end;
  
	new_menu:
	if (My_Title())
	{
	  score=0, lives=5, nb_f_squares=0; resettime =1;
	  randomize();
    new_game:
    Init_Gameset();
    New_Square();
    y_piece=58, pressed_key=0, piece=random (3)+2, joker=1;
    GraySprite8_OR(100, y_piece, 8, game_sprt[piece][0], game_sprt[piece][1], PLANE1, PLANE2);
    running=1;
    while (1)
    {
      MAT_COORD test = Attack_Position(y_piece);
      a_row = test.row;
      a_col = test.col;
      a_x_graph_pos = (test.col<<3)+2;
	 	  a_y_graph_pos = (test.row<<3)+50;
	 	  if (Is_A_Out_MAT_COORD(&test)) blinking = 0;
      else blinking = 1;
      
      //Display the score, lives and the time
      sprintf(my_buffer, "%ld", score);
      GrayDrawRect2B(112, 20, 158, 28, COLOR_WHITE, RECT_FILLED, PLANE1, PLANE2);
      My_Gray_Font(134, 20, my_buffer, 6, PLANE1, PLANE2, fnt2, 1);
      sprintf(my_buffer, "%d", lives);
      GrayDrawRect2B(112, 54, 158, 62, COLOR_WHITE, RECT_FILLED, PLANE1, PLANE2);
      My_Gray_Font(134, 54, my_buffer, 6, PLANE1, PLANE2, fnt2, 1);
      /*sprintf(buffer, "%d'%d'%d", minutes, seconds, dix);
      GrayDrawRect2B(112, 87, 158, 95, COLOR_WHITE, RECT_FILLED, PLANE1, PLANE2);
      My_Gray_Font(134, 87, buffer, 6, PLANE1, PLANE2, fnt2, 1);*/
      
      pressed_key=key(100);
      
      if (pressed_key==ESC)
      {
      	blinking=0;
      	running=0;
      	Change_HSC_Table(score); //if you've made an highscore
      	goto new_menu;
      }
      
      if (pressed_key==UP && y_piece>8)
      {
        if (blinking)
          GraySprite8_OR(a_x_graph_pos, a_y_graph_pos, 8, game_sprt[map[a_row][a_col]][0], game_sprt[map[a_row][a_col]][1], PLANE1, PLANE2);
        GraySprite8_AND(100, y_piece, 8, mask, mask, PLANE1, PLANE2);
        y_piece-=8;
        GraySprite8_OR(100, y_piece, 8, game_sprt[piece][0], game_sprt[piece][1], PLANE1, PLANE2);
      }
      
      if (pressed_key==DOWN && y_piece<90)
      {
     	  if (blinking)
     	    GraySprite8_OR(a_x_graph_pos, a_y_graph_pos, 8, game_sprt[map[a_row][a_col]][0], game_sprt[map[a_row][a_col]][1], PLANE1, PLANE2);
     	  GraySprite8_AND(100, y_piece, 8, mask, mask, PLANE1, PLANE2);
        y_piece+=8;
        GraySprite8_OR(100, y_piece, 8, game_sprt[piece][0], game_sprt[piece][1], PLANE1, PLANE2);
      }
      	
      if (pressed_key==LEFT)
      {
      	if (joker && a_row<6) {piece=map[a_row][a_col]; joker=0;}
      	StopBlinking();
      	Move_Piece(y_piece);
      	score=score+Sum_From_One_To_Kp2(add_to_score);
      }
      
      if (pressed_key==MAGIC_KEY)
      {
      	StopBlinking();
      	running=0;
      	My_Window("5 min passed");
        Change_HSC_Table(score);
        goto new_menu;
      }
      
      if (Is_A_Winner())
      {
      	StopBlinking();
      	nb_f_squares++;
      	score+=200;
      	running=0;
      	if (nb_f_squares==5 && mode==2)
      	{
      		My_Window("You finished");
      		Change_HSC_Table(score);
      		goto new_menu;
      	}
      	My_Window("Next Round");
      	goto new_game;
      }
      
      if (Is_A_Loser() && !joker)
      {
      	if (lives==0)
      	{
      		StopBlinking();
      		running=0;
      		My_Window("You lose !!!");
      		if (mode != 2) Change_HSC_Table(score); //If you have an highscore ...
      	 	goto new_menu;
      	}
      	lives--;
      	joker=1;
      }
    }
  }
  end:
  WaitForMillis(500);
  free(my_f_buffer);
  free(map);
  free(fnt1);
  free(fnt2);
  GrayOff();
  Close_HSC();
  SetIntVec (AUTO_INT_1, ai_1);
  SetIntVec (AUTO_INT_5, ai_5);
  LCD_restore(screen);
  ST_showHelp("www.cyrilmottier.com");
}

//By Cyril Mottier (orlin61@wanadoo.fr)